# coding=utf-8
from __future__ import absolute_import, print_function

from lostc import collection as lcc
from suanpan.node.base import BaseNode, Param, Port


class DockerNode(BaseNode):
    def formatNodeInfo(self, nodeInfo):
        ports = {p["uuid"]: p for p in nodeInfo.get("ports", [])}
        node = lcc.classify(
            nodeInfo.get("params", []),
            inputs=lambda p: p.get("defInPortUuid", "").startswith("in"),
            outputs=lambda p: p.get("defInPortUuid", "").startswith("out"),
            params=lambda p: "defInPortUuid" not in p,
        )
        inputs = {
            i["name"]: Port(
                uuid=i["defOutPortUuid"],
                name=i["name"],
                type=ports[i["defOutPortUuid"]]["type"],
                subtype=ports[i["defOutPortUuid"]]["subtype"],
            )
            for i in node["inputs"]
        }
        outputs = {
            o["name"]: Port(
                uuid=o["defOutPortUuid"],
                name=o["name"],
                type=ports[o["defOutPortUuid"]]["type"],
                subtype=ports[o["defOutPortUuid"]]["subtype"],
            )
            for o in node["outputs"]
        }
        params = {
            p["name"]: Param(uuid=p["controlInstUuid"], name=p["name"], type=p["type"])
            for p in node["params"]
        }
        return {"inputs": inputs, "outputs": outputs, "params": params}



# coding=utf-8
from __future__ import absolute_import, print_function

import argparse
import atexit
import base64
import functools
import itertools
import os
import re
import signal
import sys

from suanpan.arguments import Bool
from suanpan.log import logger
from suanpan.objects import Context, HasName


class HasArguments(object):
    ENVIRONMENT_ARGUMENTS_KEYWORD = "SP_PARAM"
    GLOBAL_ARGUMENTS = []
    ARGUMENTS = []

    @classmethod
    def dictFromList(cls, l):
        return dict(zip(l[0::2], l[1::2]))

    @classmethod
    def listFromDict(cls, d):
        return list(itertools.chain(*d.items()))

    @classmethod
    def getArgList(cls, *args, **kwargs):  # pylint: disable=unused-argument
        argDict = {}
        argDict.update(cls.dictFromList(cls.getArgListFromEnv()))
        argDict.update(cls.dictFromList(cls.getArgListFromCli()))
        argDict.update(cls.dictFromList(args))
        argDict.update({"--{}".format(k): v for k, v in kwargs.items()})
        return cls.listFromDict(argDict)

    @classmethod
    def getArgListFromCli(cls):
        return sys.argv[1:]

    @classmethod
    def getArgListFromEnv(cls):
        argStringBase64 = os.environ.get(cls.ENVIRONMENT_ARGUMENTS_KEYWORD, "")
        logger.debug(
            "{}(Base64)='{}'".format(cls.ENVIRONMENT_ARGUMENTS_KEYWORD, argStringBase64)
        )
        try:
            argString = base64.b64decode(argStringBase64).decode()
        except Exception:
            argString = argStringBase64  # temporary fix for SP_PARAM(Base64)
        logger.debug("{}='{}'".format(cls.ENVIRONMENT_ARGUMENTS_KEYWORD, argString))
        regex = r"(--[\w-]+)\s+(?:(?P<quote>[\"\'])(.*?)(?P=quote)|([^\'\"\s]+))"
        groups = re.findall(regex, argString, flags=re.S)
        argv = list(
            itertools.chain(*[(group[0], group[-2] or group[-1]) for group in groups])
        )
        return argv

    def loadFormatArguments(self, context, restArgs=None, **kwargs):
        arguments = self.getArguments(**kwargs)
        args, restArgs = self._parseArguments(arguments, restArgs=restArgs)
        self._loadArguments(args, arguments)
        self._formatArguments(context, arguments)
        return arguments, restArgs

    def loadCleanArguments(self, context, restArgs=None, **kwargs):
        arguments = self.getArguments(**kwargs)
        args, restArgs = self._parseArguments(arguments, restArgs=restArgs)
        self._loadArguments(args, arguments)
        self._cleanArguments(context, arguments)
        return arguments, restArgs

    def loadGlobalArguments(self, restArgs=None, **kwargs):
        arguments = self.getGlobalArguments(**kwargs)
        args, restArgs = self._parseGlobalArguments(arguments, restArgs=restArgs)
        self._loadArguments(args, arguments)
        self._formatArguments(Context(), arguments)
        return arguments, restArgs

    def loadComponentArguments(self, restArgs=None, **kwargs):
        arguments = self.getComponentArguments(**kwargs)
        args, restArgs = self._parseArguments(arguments, restArgs=restArgs)
        self._loadArguments(args, arguments)
        self._formatArguments(Context(), arguments)
        return arguments, restArgs

    def saveArguments(self, context, arguments, results):
        return {
            argument.key: argument.save(context, result)
            for argument, result in zip(arguments, results)
        }

    def getArguments(self, *args, **kwargs):
        raise NotImplementedError("Method not implemented!")

    def getGlobalArguments(self, *args, **kwargs):  # pylint: disable=unused-argument
        return self.GLOBAL_ARGUMENTS

    def getComponentArguments(self, *args, **kwargs):  # pylint: disable=unused-argument
        return self.ARGUMENTS

    @classmethod
    def _parseArguments(cls, arguments, restArgs=None, **kwargs):
        parser = argparse.ArgumentParser(allow_abbrev=False, **kwargs)
        for arg in arguments:
            arg.addParserArguments(parser)
        return parser.parse_known_args(restArgs)

    @classmethod
    def _parseGlobalArguments(cls, arguments, restArgs=None, **kwargs):
        parser = argparse.ArgumentParser(**kwargs)
        for arg in arguments:
            arg.addGlobalParserArguments(parser)
        return parser.parse_known_args(restArgs)

    @classmethod
    def _loadArguments(cls, args, arguments):
        return {arg.key: arg.load(args) for arg in arguments}

    @classmethod
    def _formatArguments(cls, context, arguments):
        return {arg.key: arg.format(context) for arg in arguments}

    @classmethod
    def _cleanArguments(cls, context, arguments):
        return {arg.key: arg.clean(context) for arg in arguments}

    @classmethod
    def defaultArgumentsFormat(cls, args, arguments):
        arguments = (arg.key.replace("-", "_") for arg in arguments)
        return {
            cls._defaultArgumentKeyFormat(arg): getattr(args, arg, None)
            for arg in arguments
        }

    @classmethod
    def _defaultArgumentKeyFormat(cls, key):
        return cls._toCamelCase(cls._removePrefix(key))

    @classmethod
    def _removePrefix(cls, string, delimiter="_", num=1):
        pieces = string.split(delimiter)
        pieces = pieces[num:] if len(pieces) > num else pieces
        return delimiter.join(pieces)

    @classmethod
    def _toCamelCase(cls, string, delimiter="_"):
        camelCaseUpper = lambda i, s: s[0].upper() + s[1:] if i and s else s
        return "".join(
            [camelCaseUpper(i, s) for i, s in enumerate(string.split(delimiter))]
        )

    @classmethod
    def argumentsDict(cls, arguments):
        result = {}
        for arg in arguments:
            keys = (arg.key, arg.alias)
            result.update({key: arg.value for key in keys if key})
        return result

    @classmethod
    def getArgumentValueFromDict(cls, data, arg):
        value = data.get(arg.alias)
        if value is not None:
            return value

        value = data.get(arg.key)
        if value is not None:
            return value

        return None

    @classmethod
    def hasArgumentValueFromDict(cls, data, arg):
        return arg.alias in data or arg.key in data


class HasLogger(HasName):
    def __init__(self):
        super(HasLogger, self).__init__()
        logger.setLogger(self.name)

    @property
    def logger(self):
        return logger


class HasDevMode(HasArguments):
    DEV_ARGUMENTS = [Bool(key="debug", default=False)]

    def getGlobalArguments(self, *args, **kwargs):
        arguments = super(HasDevMode, self).getGlobalArguments(*args, **kwargs)
        return arguments + self.DEV_ARGUMENTS


class HasInitHooks(object):
    def __init__(self, *args, **kwargs):  # pylint: disable=unused-argument
        super(HasInitHooks, self).__init__()
        self.beforeInitHooks = getattr(self, "beforeInitHooks", [])
        self.afterInitHooks = getattr(self, "afterInitHooks", [])

    def beforeInit(self, *args, **kwargs):
        pass

    def callBeforeInitHooks(self, *args, **kwargs):
        self.beforeInit(*args, **kwargs)
        for hook in self.beforeInitHooks:
            hook(*args, **kwargs)

    def addBeforeInitHooks(self, *hooks):
        self.beforeInitHooks.extend(hooks)
        return self

    def afterInit(self, *args, **kwargs):
        pass

    def callAfterInitHooks(self, *args, **kwargs):
        self.afterInit(*args, **kwargs)
        for hook in self.afterInitHooks:
            hook(*args, **kwargs)

    def addAfterInitHooks(self, *hooks):
        self.afterInitHooks.extend(hooks)
        return self


class HasSaveHooks(object):
    def __init__(self, *args, **kwargs):  # pylint: disable=unused-argument
        super(HasSaveHooks, self).__init__()
        self.beforeSaveHooks = getattr(self, "beforeSaveHooks", [])
        self.afterSaveHooks = getattr(self, "afterSaveHooks", [])

    def beforeSave(self, *args, **kwargs):
        pass

    def callBeforeSaveHooks(self, *args, **kwargs):
        self.beforeSave(*args, **kwargs)
        for hook in self.beforeSaveHooks:
            hook(*args, **kwargs)

    def addBeforeSaveHooks(self, *hooks):
        self.beforeSaveHooks.extend(hooks)
        return self

    def afterSave(self, *args, **kwargs):
        pass

    def callAfterSaveHooks(self, *args, **kwargs):
        self.afterSave(*args, **kwargs)
        for hook in self.afterSaveHooks:
            hook(*args, **kwargs)

    def addAfterSaveHooks(self, *hooks):
        self.afterSaveHooks.extend(hooks)
        return self


class HasCallHooks(object):
    def __init__(self, *args, **kwargs):  # pylint: disable=unused-argument
        super(HasCallHooks, self).__init__()
        self.beforeCallHooks = getattr(self, "beforeCallHooks", [])
        self.afterCallHooks = getattr(self, "afterCallHooks", [])

    def beforeCall(self, *args, **kwargs):
        pass

    def callBeforeCallHooks(self, *args, **kwargs):
        self.beforeCall(*args, **kwargs)
        for hook in self.beforeCallHooks:
            hook(*args, **kwargs)

    def addBeforeCallHooks(self, *hooks):
        self.beforeCallHooks.extend(hooks)
        return self

    def afterCall(self, *args, **kwargs):
        pass

    def callAfterCallHooks(self, *args, **kwargs):
        self.afterCall(*args, **kwargs)
        for hook in self.afterCallHooks:
            hook(*args, **kwargs)

    def addAfterCallHooks(self, *hooks):
        self.afterCallHooks.extend(hooks)
        return self


class HasTriggerHooks(object):
    def __init__(self, *args, **kwargs):  # pylint: disable=unused-argument
        super(HasTriggerHooks, self).__init__()
        self.beforeTriggerHooks = getattr(self, "beforeTriggerHooks", [])
        self.afterTriggerHooks = getattr(self, "afterTriggerHooks", [])

    def beforeTrigger(self, *args, **kwargs):
        pass

    def callBeforeTriggerHooks(self, *args, **kwargs):
        self.beforeTrigger(*args, **kwargs)
        for hook in self.beforeTriggerHooks:
            hook(*args, **kwargs)

    def addBeforeTriggerHooks(self, *hooks):
        self.beforeTriggerHooks.extend(hooks)
        return self

    def afterTrigger(self, *args, **kwargs):
        pass

    def callAfterTriggerHooks(self, *args, **kwargs):
        self.afterTrigger(*args, **kwargs)
        for hook in self.afterTriggerHooks:
            hook(*args, **kwargs)

    def addAfterTriggerHooks(self, *hooks):
        self.afterTriggerHooks.extend(hooks)
        return self


class HasExitHooks(object):
    def __init__(self, *args, **kwargs):  # pylint: disable=unused-argument
        super(HasExitHooks, self).__init__()
        self.beforeExitHooks = getattr(self, "beforeExitHooks", [])

    def beforeExit(self, *args, **kwargs):
        pass

    def callBeforeExitHooks(self, *args, **kwargs):
        self.beforeExit(*args, **kwargs)
        for hook in self.beforeExitHooks:
            hook(*args, **kwargs)

    def addBeforeExitHooks(self, *hooks):
        self.beforeExitHooks.extend(hooks)
        return self

    def registerBeforeExitHooks(self, *args, **kwargs):
        atexit.register(self.callBeforeExitHooks, *args, **kwargs)
        signal.signal(signal.SIGTERM, lambda signum, frame: sys.exit(1))
        return self



# coding=utf-8
from __future__ import absolute_import, print_function

from suanpan.app.base import BaseApp
from suanpan.model.arguments import HotReloadModel
from suanpan.stream import Handler, Stream
from suanpan.utils import functional


class TriggerApp(BaseApp):
    def __init__(self, streamApp, *args, **kwargs):
        super(TriggerApp, self).__init__(*args, **kwargs)
        self.streamApp = streamApp
        self.handler = None
        self.interval = None
        self.loop = None

    def __call__(self, interval=None, loop=None):
        self.interval = interval
        self.loop = self._initLoop(loop)

        def _dec(funcOrApp):
            funcOrApp = (
                self.handler
                if isinstance(funcOrApp, TriggerApp)
                else functional.instancemethod(funcOrApp)
            )
            self.handler = Handler.use(funcOrApp)
            return self

        return _dec

    def _initLoop(self, loop):
        if loop is None:
            return None
        if callable(loop):
            return functional.instancemethod(loop)
        return lambda obj: loop

    def input(self, argument):
        if self.streamApp.isComponentArgument(argument):
            return self.componentParam(argument)

        def _dec(funcOrApp):
            funcOrApp = (
                self.handler
                if isinstance(funcOrApp, TriggerApp)
                else functional.instancemethod(funcOrApp)
            )
            self.handler = Handler.input(argument)(funcOrApp)
            return self

        return _dec

    def output(self, argument):
        if isinstance(argument, HotReloadModel):
            raise Exception("{} can't be set as output!".format(argument.name))

        def _dec(funcOrApp):
            funcOrApp = (
                self.handler
                if isinstance(funcOrApp, TriggerApp)
                else functional.instancemethod(funcOrApp)
            )
            self.handler = Handler.output(argument)(funcOrApp)
            return self

        return _dec

    def param(self, argument):
        if self.streamApp.isComponentArgument(argument):
            return self.componentParam(argument)

        self.streamApp.globalArguments.append(argument)

        def _dec(funcOrApp):
            funcOrApp = (
                self.handler
                if isinstance(funcOrApp, TriggerApp)
                else functional.instancemethod(funcOrApp)
            )
            self.handler = Handler.use(funcOrApp)
            return self

        return _dec

    def column(self, argument):
        if self.streamApp.isComponentArgument(argument):
            return self.componentParam(argument)

        self.streamApp.globalArguments.append(argument)

        def _dec(funcOrApp):
            funcOrApp = (
                self.handler
                if isinstance(funcOrApp, TriggerApp)
                else functional.instancemethod(funcOrApp)
            )
            self.handler = Handler.use(funcOrApp)
            return self

        return _dec

    def componentParam(self, argument):
        self.streamApp.arguments.append(argument)

        def _dec(funcOrApp):
            funcOrApp = (
                self.handler
                if isinstance(funcOrApp, TriggerApp)
                else functional.instancemethod(funcOrApp)
            )
            self.handler = Handler.use(funcOrApp)
            return self

        return _dec

    def send(self, *args, **kwargs):
        self.streamApp.stream.send(*args, **kwargs)


class StreamApp(BaseApp):
    COMPONENT_ARGUMENT_CLASSES = (HotReloadModel,)

    def __init__(self, *args, **kwargs):
        super(StreamApp, self).__init__(*args, **kwargs)
        self.triggerApp = TriggerApp(self)
        self.stream = None
        self.handler = None
        self.arguments = []
        self.globalArguments = []

    def __call__(self, *args, **kwargs):
        attrs = {
            "GLOBAL_ARGUMENTS": self.globalArguments,
            "ARGUMENTS": self.arguments,
            "INTERVAL": self.trigger.interval,
            "loop": self.trigger.loop,
            "call": self.handler,
            "trigger": self.trigger.handler,
            "beforeInitHooks": self.beforeInitHooks + self.trigger.beforeInitHooks,
            "afterInitHooks": self.afterInitHooks + self.trigger.afterInitHooks,
            "beforeExitHooks": self.beforeExitHooks + self.trigger.beforeExitHooks,
            "beforeCallHooks": self.beforeCallHooks,
            "afterCallHooks": self.afterCallHooks,
            "beforeTriggerHooks": self.trigger.beforeCallHooks,
            "afterTriggerHooks": self.trigger.afterCallHooks,
        }
        attrs = {k: v for k, v in attrs.items() if v is not None}
        ADSteam = type("ADSteam", (Stream,), attrs)
        self.stream = ADSteam(*args, **kwargs)
        return self

    @property
    def trigger(self):
        return self.triggerApp

    def start(self, *args, **kwargs):
        if not self.stream:
            raise Exception("{} is not ready".format(self.name))
        self.stream.start(*args, **kwargs)
        return self

    def input(self, argument):
        if self.isComponentArgument(argument):
            return self.componentParam(argument)

        def _dec(funcOrApp):
            funcOrApp = (
                self.handler
                if isinstance(funcOrApp, StreamApp)
                else functional.instancemethod(funcOrApp)
            )
            self.handler = Handler.input(argument)(funcOrApp)
            return self

        return _dec

    def output(self, argument):
        if isinstance(argument, HotReloadModel):
            raise Exception("{} can't be set as output!".format(argument.name))

        def _dec(funcOrApp):
            funcOrApp = (
                self.handler
                if isinstance(funcOrApp, StreamApp)
                else functional.instancemethod(funcOrApp)
            )
            self.handler = Handler.output(argument)(funcOrApp)
            return self

        return _dec

    def param(self, argument):
        if self.isComponentArgument(argument):
            return self.componentParam(argument)

        self.globalArguments.append(argument)

        def _dec(funcOrApp):
            funcOrApp = (
                self.handler
                if isinstance(funcOrApp, StreamApp)
                else functional.instancemethod(funcOrApp)
            )
            self.handler = Handler.use(funcOrApp)
            return self

        return _dec

    def column(self, argument):
        if self.isComponentArgument(argument):
            return self.componentParam(argument)

        self.globalArguments.append(argument)

        def _dec(funcOrApp):
            funcOrApp = (
                self.handler
                if isinstance(funcOrApp, StreamApp)
                else functional.instancemethod(funcOrApp)
            )
            self.handler = Handler.use(funcOrApp)
            return self

        return _dec

    def componentParam(self, argument):
        self.arguments.append(argument)

        def _dec(funcOrApp):
            funcOrApp = (
                self.handler
                if isinstance(funcOrApp, StreamApp)
                else functional.instancemethod(funcOrApp)
            )
            self.handler = Handler.use(funcOrApp)
            return self

        return _dec

    def isComponentArgument(self, argument):
        return isinstance(argument, self.COMPONENT_ARGUMENT_CLASSES)

    def send(self, *args, **kwargs):
        self.stream.send(*args, **kwargs)



# coding=utf-8
from __future__ import absolute_import, print_function

from suanpan.arguments import Arg, String
from suanpan.dw import dw


class DWArg(Arg):
    def getOutputTmpValue(self, *args):
        return "_".join(args)


class Table(DWArg):
    def __init__(self, key, table, partition, **kwargs):
        kwargs.setdefault("required", False)
        super(Table, self).__init__(key)
        self.table = String(key=table, required=kwargs["required"])
        self.partition = String(key=partition)

    @property
    def isSet(self):
        return True

    def addParserArguments(self, parser):
        self.table.addParserArguments(parser)
        self.partition.addParserArguments(parser)

    def load(self, args):
        self.table.load(args)
        self.partition.load(args)
        if self.table.value:
            self.value = {"table": self.table.value, "partition": self.partition.value}

    def format(self, context):
        if self.value:
            self.value = dw.readTable(self.table.value, self.partition.value)
        return self.value

    def save(self, context, result):
        data = result.value
        dw.writeTable(self.table.value, data)
        self.logSaved(self.table.value)
        return self.value

    def getOutputTmpArg(self, *args):
        value = self.getOutputTmpValue(  # pylint: disable=assignment-from-no-return
            *args
        )
        return ("--{}".format(self.table.key), value) if value is not None else tuple()



# coding=utf-8
from __future__ import absolute_import, print_function

from suanpan.node import node
from suanpan.arguments.auto import AutoArg


class DataFrame(AutoArg):
    MAPPING = {
        "data": "suanpan.storage.arguments.Csv",
        "table": "suanpan.dw.arguments.Table",
    }

    def __init__(self, key, table, partition, *args, **kwargs):
        port = node.get(table) or node.get(key)
        if not port:
            raise Exception(
                "No such {}: {}".format(self.name, ",".join([key, table, partition]))
            )
        kwargs.setdefault(self.TYPE_KEY, port.type)
        kwargs.update(key=key, table=table, partition=partition)
        super(DataFrame, self).__init__(*args, **kwargs)



# coding=utf-8
from __future__ import absolute_import, print_function

from suanpan.screenshots.base import ScreenshotsIndexSaver
from suanpan.state.arguments import StorageSaverArg


class Screenshots(StorageSaverArg):
    STATE_CLASS = ScreenshotsIndexSaver
    STATE_PATTERN = "{index}.png"



# coding=utf-8
from __future__ import absolute_import, print_function

import pandas
from odps import ODPS
from odps.df import DataFrame
from odps.df.expr.expressions import CollectionExpr

from suanpan.log import logger


class OdpsDataWarehouse(object):
    DEFAULT_ODPS_PROJECT_NAME = "None"

    DTYPE_SQLTYPE_MAPPING = {
        "int8": "int",
        "int16": "int",
        "int32": "int",
        "int64": "bigint",
        "float32": "double",
        "float64": "double",
        "class": "string",
        "datetime": "datetime",
    }

    def __init__(
        self,  # pylint: disable=unused-argument
        odpsAccessId,
        odpsAccessKey,
        odpsEndpoint,
        odpsProject,
        **kwargs
    ):
        self.odpsAccessId = odpsAccessId
        self.odpsAccessKey = odpsAccessKey
        self.odpsEndpoint = odpsEndpoint
        self.odpsProject = odpsProject

        self._connection = self.connect()

    def connect(self):
        connection = ODPS(
            access_id=self.odpsAccessId,
            secret_access_key=self.odpsAccessKey,
            endpoint=self.odpsEndpoint,
            project=self.odpsProject,
        )
        return connection

    def testConnection(self, connection):
        connection.execute_sql("select current_timestamp()")
        logger.info("Odps connected...")

    def readTable(
        self,
        table,
        partition=None,
        database=DEFAULT_ODPS_PROJECT_NAME,  # pylint: disable=unused-argument
    ):
        df = (
            DataFrame(self._connection.get_table(table).get_partition(partition))
            if partition
            else DataFrame(self._connection.get_table(table))
        )

        if df.count() == 0:
            return self._createEmptyPandasDF(self.getColumnsFromOdps(df))

        return df.to_pandas()

    def writeTable(
        self,
        table,
        data,
        partition=None,
        database=DEFAULT_ODPS_PROJECT_NAME,  # pylint: disable=unused-argument
        overwrite=True,
    ):
        isOdpsDataFrame = isinstance(data, CollectionExpr)
        if isOdpsDataFrame:
            self._persistData(table, data, partition, database, overwrite)
        else:
            if data.size == 0:
                columns = self.getColumns(data)
                self.dropTable(table)
                self.createTable(table, columns)
            else:
                data = DataFrame(data)
                self._persistData(table, data, partition, database, overwrite)

    def createTable(self, table, columns):
        self.execute(
            u"create table if not exists {table} ({columns})".format(
                table=table,
                columns=u",".join(
                    [
                        u"{} {}".format(column["name"], column["type"])
                        for column in columns
                    ]
                ),
            )
        )

    def dropTable(self, table):
        self._connection.delete_table(table, if_exists=True)

    def execute(self, sql):
        try:
            self._connection.execute_sql(sql)
        except Exception as e:
            logger.error("SQL: {}".format(sql))
            raise e

    def getColumns(self, data):
        return [
            {"name": name.split(".")[-1], "type": self.toSqlType(dtype)}
            for name, dtype in data.dtypes.items()
        ]

    def getColumnsFromOdps(self, df):
        return [
            {"name": column.name.split(".")[-1], "type": str(column.type).lower()}
            for column in df.schema.columns
        ]

    def toSqlType(self, dtype):
        typeName = self.shortenDatetimeType(dtype.name)
        return self.DTYPE_SQLTYPE_MAPPING.get(typeName, "string")

    def shortenDatetimeType(self, typeName):
        datatimeType = "datetime"
        return datatimeType if typeName.startswith(datatimeType) else typeName

    def _persistData(
        self,
        table,
        data,
        partition=None,
        database=DEFAULT_ODPS_PROJECT_NAME,  # pylint: disable=unused-argument
        overwrite=True,
    ):
        if partition:
            data.persist(
                table,
                partition=partition,
                drop_partition=overwrite,
                create_partition=True,
                odps=self._connection,
            )
        else:
            data.persist(
                table,
                partition=partition,
                drop_table=overwrite,
                create_table=True,
                odps=self._connection,
            )

    def _createEmptyPandasDF(self, columns):
        columnNames = [column["name"] for column in columns]
        df = pandas.DataFrame(columns=columnNames)
        for column in columns:
            name = column["name"]
            dType = column["type"]

            if dType == "string":
                df[name] = df[name].astype("object")
            else:
                df[name] = df[name].astype(dType)
        return df



# coding=utf-8
from __future__ import absolute_import, print_function

from suanpan import path
from suanpan.arguments import Arg
from suanpan.components import Result
from suanpan.storage import storage
from suanpan.utils import csv, image, json, npy, text


class StorageArg(Arg):
    def getOutputTmpValue(self, *args):
        return storage.delimiter.join(args)


class File(StorageArg):
    FILENAME = "file"
    FILETYPE = None

    def __init__(self, key, **kwargs):
        fileName = kwargs.pop("name", self.FILENAME)
        fileType = kwargs.pop("type", self.FILETYPE)
        self.fileName = (
            "{}.{}".format(fileName, fileType.lower()) if fileType else fileName
        )
        self.objectPrefix = None
        self.objectName = None
        self.filePath = None
        super(File, self).__init__(key, **kwargs)

    @property
    def isSet(self):
        return True

    def load(self, args):
        self.objectPrefix = super(File, self).load(args)
        if self.objectPrefix:
            self.objectName = storage.storagePathJoin(self.objectPrefix, self.fileName)
            self.filePath = storage.getPathInTempStore(self.objectName)
            path.safeMkdirsForFile(self.filePath)
        self.value = self.filePath
        return self.filePath

    def format(self, context):
        if self.filePath:
            storage.download(self.objectName, self.filePath)
        return self.filePath

    def clean(self, context):
        if self.filePath:
            path.remove(self.filePath)
        return self.filePath

    def save(self, context, result):
        filePath = result.value
        storage.upload(self.objectName, filePath)
        self.logSaved(self.objectName)
        return self.objectPrefix


class Folder(StorageArg):
    def __init__(self, key, **kwargs):
        super(Folder, self).__init__(key, **kwargs)
        self.folderName = None
        self.folderPath = None

    @property
    def isSet(self):
        return True

    def load(self, args):
        self.folderName = super(Folder, self).load(args)
        if self.folderName:
            self.folderPath = storage.getPathInTempStore(self.folderName)
        if self.folderPath:
            path.safeMkdirs(self.folderPath)
        self.value = self.folderPath
        return self.folderPath

    def format(self, context):
        if self.folderPath:
            storage.download(self.folderName, self.folderPath)
        return self.folderPath

    def clean(self, context):
        if self.folderPath:
            path.empty(self.folderPath)
        return self.folderPath

    def save(self, context, result):
        folderPath = result.value
        storage.upload(self.folderName, folderPath)
        self.logSaved(self.folderName)
        return self.folderName


class Data(File):
    FILENAME = "data"


class Json(Data):
    FILETYPE = "json"

    def format(self, context):
        super(Json, self).format(context)
        if self.filePath:
            self.value = json.load(self.filePath)
        return self.value

    def save(self, context, result):
        json.dump(result.value, self.filePath)
        return super(Json, self).save(context, Result.froms(value=self.filePath))


class Csv(Data):
    FILETYPE = "csv"

    def format(self, context):
        super(Csv, self).format(context)
        if self.filePath:
            self.value = csv.load(self.filePath)
        return self.value

    def save(self, context, result):
        csv.dump(result.value, self.filePath)
        return super(Csv, self).save(context, Result.froms(value=self.filePath))


class Table(Csv):
    pass


class Npy(Data):
    FILETYPE = "npy"

    def format(self, context):
        super(Npy, self).format(context)
        if self.filePath:
            self.value = npy.load(self.filePath)
        return self.value

    def save(self, context, result):
        npy.dump(result.value, self.filePath)
        return super(Npy, self).save(context, Result.froms(value=self.filePath))


class String(Data):
    FILETYPE = "txt"

    def format(self, context):
        super(String, self).format(context)
        if self.filePath:
            self.value = text.load(self.filePath)
        return self.value

    def save(self, context, result):
        text.dump(result.value, self.filePath)
        return super(String, self).save(context, Result.froms(value=self.filePath))


class Visual(String):
    FILENAME = "part-00000"
    FILETYPE = ""


class Model(File):
    FILENAME = "model"


class H5Model(Model):
    FILETYPE = "h5"


class Checkpoint(Model):
    FILETYPE = "ckpt"


class JsonModel(Model):
    FILETYPE = "json"


class Image(File):
    FILENAME = "image"
    FILETYPE = "png"

    def format(self, context):
        super(Image, self).format(context)
        if self.filePath:
            self.value = image.read(self.filePath)
        return self.value

    def save(self, context, result):
        image.save(result.value, self.filePath)
        return super(Image, self).save(context, Result.froms(value=self.filePath))


class Screenshots(Image):
    FILENAME = "screenshots"



# coding=utf-8
from __future__ import absolute_import, print_function

from suanpan import arguments as base
from suanpan.arguments import auto  # pylint: disable=unused-import
from suanpan.dw import arguments as dw
from suanpan.model import arguments as model
from suanpan.mq import arguments as mq  # pylint: disable=unused-import
from suanpan.mstorage import arguments as mstorage
from suanpan.node import arguments as node
from suanpan.storage import arguments as storage

Int = base.Int
Float = base.Float
Bool = base.Bool
String = base.String
List = base.List
ListOfString = base.ListOfString
ListOfInt = base.ListOfInt
ListOfFloat = base.ListOfFloat
ListOfBool = base.ListOfBool
IntOrFloat = base.IntOrFloat
IntFloatOrString = base.IntFloatOrString
BoolOrString = base.BoolOrString
StringOrListOfFloat = base.StringOrListOfFloat
Json = base.Json

File = storage.File
Folder = storage.Folder
Data = storage.Data
Csv = storage.Csv
Npy = storage.Npy
Visual = storage.Visual
Model = model.HotReloadModel
H5Model = storage.H5Model
Checkpoint = storage.Checkpoint
JsonModel = storage.JsonModel
Image = storage.Image

Table = dw.Table
DataFrame = node.DataFrame



# coding=utf-8
from __future__ import absolute_import, print_function

from suanpan.node.docker import DockerNode


class StreamNode(DockerNode):
    def formatNodeInfo(self, nodeInfo):
        nodeInfo["params"] = [
            {"name": name, **param}
            for name, param in nodeInfo.get("params", {}).items()
        ]
        return super(StreamNode, self).formatNodeInfo(nodeInfo)



# coding=utf-8
from __future__ import absolute_import, print_function

import base64
import copy
import itertools
import os
import re
from collections import namedtuple

from lostc import collection as lcc

from suanpan.log import logger
from suanpan.objects import HasName
from suanpan.utils import json

Param = namedtuple("Param", ["uuid", "name", "type"])
Port = namedtuple("Port", ["uuid", "name", "type", "subtype"])


class BaseNode(HasName):
    NODE_INFO_KEY = "SP_NODE_INFO"
    DEFAULT_NODE_INFO = {"info": {}, "inputs": {}, "outputs": {}, "params": {}}

    def __init__(self):
        super(BaseNode, self).__init__()
        self._node = self.load()

    def __getattr__(self, key):
        value = self._getPort(key) if re.match(r"(in|out)\d+", key) else self.get(key)
        if value is not None:
            raise Exception("{} not contains {}".format(self.name, key))
        return value

    def _getPort(self, key):
        return lcc.first(
            itertools.chain(self.inputs, self.outputs), lambda i: i.uuid == key
        )

    def get(self, key, default=None):
        for _, collection in self._node.items():
            if isinstance(collection, dict):
                if key in collection:
                    return collection[key]
        return default

    @property
    def info(self):
        return self._node["info"]

    @property
    def inputs(self):
        return self._node["inputs"].values()

    @property
    def ins(self):
        return self._node["inputs"].values()

    @property
    def outputs(self):
        return self._node["outputs"].values()

    @property
    def outs(self):
        return self._node["outputs"].values()

    @property
    def params(self):
        return self._node["params"].values()

    def loadFromEnv(self):
        nodeInfoBase64 = os.environ.get(self.NODE_INFO_KEY, "")
        logger.debug("{}(Base64)='{}'".format(self.NODE_INFO_KEY, nodeInfoBase64))
        nodeInfoString = base64.b64decode(nodeInfoBase64).decode()
        nodeInfo = json.loads(nodeInfoString)
        return nodeInfo

    def formatNodeInfo(self, nodeInfo):
        raise NotImplementedError("Method not implemented!")

    def defaultNodeInfo(self):
        return copy.deepcopy(self.DEFAULT_NODE_INFO)

    def _updateInfo(self, *infos):
        result = self.defaultNodeInfo()
        keys = result.keys()
        for info in infos:
            for key in keys:
                result[key].update(info[key])
        return result

    def load(self):
        return self._updateInfo(self.formatNodeInfo(self.loadFromEnv()))
